/*
 * Copyright 2022 KylinSoft Co., Ltd.
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "multitaskviewmodel.h"
#ifdef KWINEFFECT
    #include "ukui-window-switch-kwineffect/multitaskviewmodelbycompositehandler.h"
#else
    #include "ukui-window-switch-wlcom/multitaskviewmodelbywlcomhandler.h"
#endif

#include <algorithm>

#include <QGSettings>
#include <QSettings>
#include <QGuiApplication>
#include <QScreen>
#include <QDBusConnection>
#include <QDBusMessage>
#include <QFile>
#include <QJsonDocument>
#include <QJsonObject>

static const QString UKUI_STYLE_SCHEMA = "org.ukui.style";
static const QString UKUI_STYLE_SCHEMA_STYLENAME = "styleName";
static const QString UKUI_STYLE_SCHEMA_FONTSIZE = "systemFontSize";
static const QString UKUI_STYLE_SCHEMA_FONT = "systemFont";
static const QString UKUI_STYLE_SCHEMA_WIDGETNAME = "widgetThemeName";
const QString IGNORE_WINDOW_CONFIG = "/usr/share/ukui/ukui-panel/panel-commission.ini";

MultitaskViewModel::MultitaskViewModel(QObject *parent) : QObject(parent)
{
    //kwineffect相关接口无法在非effect插件的环境下编译通过，需要通过预编译来进行判断
#ifdef KWINEFFECT
    model_handler = new MultitaskViewModelByCompositeHandler(this);
#else
    model_handler = new MultitaskViewModelByWlcomHandler(this);
    connect(model_handler, &MultitaskViewModelByWlcomHandler::desktopRemoved, this, &MultitaskViewModel::desktopRemoved);
#endif
    updateModelData();
    updateInputMode();
    connectSignals();
    if (QGSettings::isSchemaInstalled(UKUI_STYLE_SCHEMA.toUtf8())) {
        m_settings = new QGSettings(UKUI_STYLE_SCHEMA.toUtf8());
    }
}

MultitaskViewModel::~MultitaskViewModel()
{
    delete model_handler;
}

bool MultitaskViewModel::currentDesktopHasWindow() const
{
    for (int screenIndex = 0; screenIndex < logicalScreenCount(); screenIndex++) {
        qDebug()<<"getAppWindowList(getCurrentDesktop(), screenIndex).size()"<<getAppWindowList(getCurrentDesktop(), screenIndex).size();
        if (getAppWindowList(getCurrentDesktop(), screenIndex).size() > 0) {
            return true;
        }
    }
    return false;
}

QList<QVariant> MultitaskViewModel::getAppWindowList(int desktopIndex, int screenIndex) const
{
    if (!validateDesktopIndex(desktopIndex) ||!validateScreenIndex(screenIndex)) {
        return QList<QVariant>();
    }

    auto desktopIter = std::find_if(m_desktopList.cbegin(),
                                    m_desktopList.cend(),
                                    [desktopIndex](const QVariant& desktop) {
        return desktop.value<Desktop>().index() == desktopIndex;
    });
    if (desktopIter == m_desktopList.cend()) {
        return QList<QVariant>();
    }

    const auto& desktop = desktopIter->value<Desktop>();
    const auto& screenList = desktop.screenList();
    auto screenIter = std::find_if(screenList.cbegin(),
                                   screenList.cend(),
                                   [screenIndex](const QVariant& screen) {
        return screen.value<Screen>().index() == screenIndex;
    });
    if (screenIter == screenList.cend()) {
        return QList<QVariant>();
    }

    const auto& screen = screenIter->value<Screen>();
    auto appWindowList = screen.windowList();
    std::reverse(appWindowList.begin(), appWindowList.end());
    return appWindowList;
}


QList<QVariant> MultitaskViewModel::getAllAppWindowList() const
{
    QList<QVariant> allAppWindowList;
    for(int desktopIndex = 1; desktopIndex <= m_desktopList.size(); desktopIndex++) {
        for (int screenIndex = 0; screenIndex < logicalScreenCount(); screenIndex++) {
             allAppWindowList += getAppWindowList(desktopIndex, screenIndex);
        }
    }
    return allAppWindowList;
}

QString MultitaskViewModel::getDesktopUuid(int desktopIndex)
{
    return model_handler->getDesktopUuid(desktopIndex);
}

QString MultitaskViewModel::getOutputUuid(int outputIndex)
{
    return model_handler->getOutputUuid(outputIndex);
}

void MultitaskViewModel::removeAppWindow(QString windowId)
{
    model_handler->removeAppWindow(windowId);
}

void MultitaskViewModel::removeAllAppWindowOnDesktop(int desktopIndex, int screenIndex)
{
    QList<QVariant> windowList = getAppWindowList(desktopIndex, screenIndex);
    for (QVariant &windowVariant : windowList) {
        Window window = windowVariant.value<Window>();
        removeAppWindow(window.uuid());
    }
}

void MultitaskViewModel::moveAppWindowToDesktop(QString windowId,
                                                int srcDesktopInex,
                                                int destDesktopInex) const
{
    if (!validateDesktopIndex(srcDesktopInex) || !validateDesktopIndex(destDesktopInex)) {
        return;
    }
    model_handler->moveAppWindowToDesktop(windowId, srcDesktopInex, destDesktopInex);
}

void MultitaskViewModel::activateAppWindow(QString windowId)
{
    model_handler->activateAppWindow(windowId);
}

QIcon MultitaskViewModel::getWindowIcon(QString windowId)
{
    return model_handler->getWindowIcon(windowId);
}

QList<QVariant> MultitaskViewModel::getDesktopList() const
{
    return m_desktopList;
}

void MultitaskViewModel::removeDesktop(int desktopIndex)
{
    auto desktopCount = model_handler->getNumberOfDesktops();
    if (desktopCount <= 1) {
        return;
    }
    if (!validateDesktopIndex(desktopIndex)) {
        return;
    }
    const auto oldDesktopIndex = desktopIndex;
    model_handler->removeDesktop(desktopIndex);
    updateModelData();
#ifdef KWINEFFECT
    emit desktopRemoved(oldDesktopIndex);
#endif
}

void MultitaskViewModel::appendDesktop()
{
    auto desktopCount = model_handler->getNumberOfDesktops();
    if (desktopCount >= MAX_DESKTOP_COUNT) {
        return;
    }
    model_handler->appendDesktop();
    updateModelData();
    emit desktopAppended(desktopCount + 1);
}

void MultitaskViewModel::moveDesktop(int srcDesktopIndex, int destDesktopIndex)
{
    if (!validateDesktopIndex(srcDesktopIndex)
            || !validateDesktopIndex(destDesktopIndex)) {
        return;
    }
    if (srcDesktopIndex == destDesktopIndex) {
        return;
    }
    model_handler->moveDesktop(srcDesktopIndex, destDesktopIndex);
    updateModelData();
    emit desktopMoved(srcDesktopIndex, destDesktopIndex);
}

int MultitaskViewModel::getCurrentDesktop() const
{
    return model_handler->getCurrentDesktop();
}

bool MultitaskViewModel::isCurrentDesktopShowingDesktop() const
{
    return model_handler->isCurrentDesktopShowingDesktop();
}

void MultitaskViewModel::setCurrentDesktop(int desktopIndex)
{
    if (!validateDesktopIndex(desktopIndex)) {
        return;
    }
    model_handler->setCurrentDesktop(desktopIndex);
}

QRect MultitaskViewModel::getFullScreenGeometry(int desktopIndex, int screenIndex) const
{
    return model_handler->getFullScreenGeometry(desktopIndex, screenIndex);
}

bool MultitaskViewModel::hasModalChildWindow(QString windowId)
{
    return model_handler->hasModalChildWindow(windowId);
}

QString MultitaskViewModel::getUKUIStyleName() const
{
    // default to ukui-default
    QString styleName = "ukui-default";
    if (m_settings && m_settings->keys().contains(UKUI_STYLE_SCHEMA_STYLENAME)) {
        styleName = m_settings->get(UKUI_STYLE_SCHEMA_STYLENAME).toString();
    } else {
        qWarning() << QString("未查询到Gesetting的键值(%1)").arg(UKUI_STYLE_SCHEMA_STYLENAME);
    }
    return styleName;
}

double MultitaskViewModel::getUKUIFontSize() const
{
    // default to 11
    double fontSize = 11;
    if (m_settings && m_settings->keys().contains(UKUI_STYLE_SCHEMA_FONTSIZE)) {
        fontSize = m_settings->get(UKUI_STYLE_SCHEMA_FONTSIZE).toDouble();
    } else {
        qWarning() << QString("未查询到Gesetting的键值(%1)").arg(UKUI_STYLE_SCHEMA_FONTSIZE);
    }
    return fontSize;
}

QString MultitaskViewModel::getUKUIFont() const
{
    // default to 11
    QString font = "Noto Sans CJK SC";
    if (m_settings && m_settings->keys().contains(UKUI_STYLE_SCHEMA_FONT)) {
        font = m_settings->get(UKUI_STYLE_SCHEMA_FONT).toString();
    } else {
        qWarning() << QString("未查询到Gesetting的键值(%1)").arg(UKUI_STYLE_SCHEMA_FONT);
    }
    return font;
}

void MultitaskViewModel::setIsOpenGLBackend(bool isOpenGLBackend)
{
    m_isOpenGLBackend = isOpenGLBackend;
}

int MultitaskViewModel::getRadius(QString type) const
{
    return m_isOpenGLBackend ? getUKUIRadius(type) : 0;
}

int MultitaskViewModel::getAllWindowCount() const
{
    return getAllAppWindowList().size();
}

int MultitaskViewModel::getUKUIRadius(QString type) const
{
    // default to 8
    QString widgetName = "default";
    if (m_settings && m_settings->keys().contains(UKUI_STYLE_SCHEMA_WIDGETNAME)) {
        widgetName = m_settings->get(UKUI_STYLE_SCHEMA_WIDGETNAME).toString();
    } else {
        qWarning() << QString("未查询到Gesetting的键值(%1)").arg(UKUI_STYLE_SCHEMA_WIDGETNAME);
        return 8;
    }

    QFile file(QString("/usr/share/qt5-ukui-platformtheme/themeconfig/%1.json").arg(widgetName));
    if (!file.open(QIODevice::ReadOnly)) {
        qWarning() << "无法打开JSON文件";
        return 8;
    }

    QByteArray jsonData = file.readAll();
    file.close();

    // 将JSON数据转换为QJsonDocument对象
    QJsonParseError error;
    QJsonDocument jsonDoc = QJsonDocument::fromJson(jsonData, &error);
    if (error.error != QJsonParseError::NoError) {
        qWarning() << "JSON格式错误:" << error.errorString();
        return 8;
    }

    // 获取属性
    QJsonObject rootObj = jsonDoc.object();
    QJsonObject radiusObj = rootObj["Radius"].toObject();
    QJsonObject radiusValueObj = radiusObj[type].toObject();

    return radiusValueObj["value"].toString().toInt();
}

void MultitaskViewModel::onWindowAdded()
{
    updateModelData();
    emit appWindowCreated();
}

void MultitaskViewModel::onWindowDeleted(QString windowId)
{
    updateModelData();
    emit appWindowRemoved(windowId);
}

void MultitaskViewModel::onWindowActivated()
{
    emit appWindowActivated();
}

void MultitaskViewModel::onDesktopPresenceChanged(QString windowId,
                                                  int oldDesktopIndex, int newDesktopIndex)
{
    updateModelData();
    emit appWindowDesktopChanged(windowId, oldDesktopIndex, newDesktopIndex);
}

void MultitaskViewModel::onNumberScreensChanged()
{
    updateModelData();
    emit screenCountChanged();
}

void MultitaskViewModel::onCurrentDesktopChanged(int oldCurrentDesktopIndex, int newCurrentDesktopIndex)
{
    updateModelData();
    emit currentDesktopChanged(oldCurrentDesktopIndex, newCurrentDesktopIndex);
}

bool MultitaskViewModel::isExtensionMode()
{
    return model_handler->isExtensionMode();
}

int MultitaskViewModel::logicalScreenCount() const
{
    return model_handler->logicalScreenCount();
}

bool MultitaskViewModel::isTabletMode() const
{
    // default to PC mode.
    bool tabletMode = false;
    QDBusMessage message = QDBusMessage::createMethodCall("com.kylin.statusmanager.interface",
                                                          "/",
                                                          "com.kylin.statusmanager.interface",
                                                          "get_current_tabletmode");
    QDBusMessage response = QDBusConnection::sessionBus().call(message);
    if (response.type() == QDBusMessage::ReplyMessage) {
        tabletMode = response.arguments().takeFirst().toBool();
    } else {
        qDebug() << "get tablet mode failed " << response.type();
    }
    return tabletMode;
}

QString MultitaskViewModel::getCurrentLanguage() const
{
    // default to "Chinese".
    QLocale locale = QLocale::system();
    QString language = "Chinese";
    if (locale.language() == QLocale::Chinese) {
        language = "Chinese";
    } else if (locale.language() == QLocale::Tibetan) {
        language = "Tibetan";
    }
    return language;
}

void MultitaskViewModel::updateModelData()
{
    QList<QVariant> desktopList;
    const auto desktopCount = model_handler->getNumberOfDesktops();
    for (int desktopIndex = 1; desktopIndex <= desktopCount; desktopIndex++) {
        QList<QVariant> screenList;
        for (int screenIndex = 0; screenIndex < logicalScreenCount(); screenIndex++) {
            auto screen = QVariant::fromValue(Screen(screenIndex, model_handler->createWindowList(desktopIndex, screenIndex)));
            screenList.append(std::move(screen));
        }
        auto desktop = QVariant::fromValue(Desktop(desktopIndex, screenList));
        desktopList.append(std::move(desktop));
    }
    m_desktopList = std::move(desktopList);
}

void MultitaskViewModel::updateInputMode()
{

    QDBusMessage message = QDBusMessage::createMethodCall("com.kylin.statusmanager.interface",
                                                          "/",
                                                          "com.kylin.statusmanager.interface",
                                                          "get_inputmethod_mode");
    QDBusMessage response = QDBusConnection::sessionBus().call(message);
    if (response.type() == QDBusMessage::ReplyMessage) {
        m_inputMode = response.arguments().takeFirst().toBool();
        qDebug() << "m_inputMode =" << m_inputMode;
    } else {
        // invalid response, keep default.
    }
}

// static
bool MultitaskViewModel::validateDesktopIndex(int desktopIndex)
{
    return desktopIndex >= 1 && desktopIndex <= MAX_DESKTOP_COUNT;
}

bool MultitaskViewModel::validateScreenIndex(int screenIndex) const
{
    return screenIndex >= 0 && screenIndex < logicalScreenCount();
}

void MultitaskViewModel::connectSignals()
{
    QObject::connect(model_handler, &MultitaskViewModelHandler::appWindowCreated,
                     this, &MultitaskViewModel::onWindowAdded);
    QObject::connect(model_handler, &MultitaskViewModelHandler::appWindowRemoved,
                     this, &MultitaskViewModel::onWindowDeleted);
    QObject::connect(model_handler, &MultitaskViewModelHandler::appWindowActivated,
                     this, &MultitaskViewModel::onWindowActivated);
    QObject::connect(model_handler, &MultitaskViewModelHandler::appWindowDesktopChanged,
                     this, &MultitaskViewModel::onDesktopPresenceChanged);
    QObject::connect(model_handler, &MultitaskViewModelHandler::currentDesktopChanged,
                     this, &MultitaskViewModel::onCurrentDesktopChanged);
    QObject::connect(model_handler, &MultitaskViewModelHandler::tabletModeChanged,
                     this, &MultitaskViewModel::tabletModeChanged);
    QObject::connect(model_handler, &MultitaskViewModelHandler::screenSizeChanged,
                     this, &MultitaskViewModel::screenSizeChanged);
    QObject::connect(qGuiApp, &QGuiApplication::screenAdded,
                     this, &MultitaskViewModel::onNumberScreensChanged);
    QObject::connect(qGuiApp, &QGuiApplication::screenRemoved,
                     this, &MultitaskViewModel::onNumberScreensChanged);
    QDBusConnection::sessionBus().connect("com.kylin.statusmanager.interface",
                                          "/",
                                          "com.kylin.statusmanager.interface",
                                          "inputmethod_change_signal",
                                          this,
                                          SLOT(inputModeSlot(bool)));
}

void MultitaskViewModel::inputModeSlot(bool inputMode)
{
    m_inputMode = inputMode;
    emit inputModeChanged(m_inputMode);
}
