/*
 * Copyright 2022 KylinSoft Co., Ltd.
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#ifndef MULTITASKVIEWMODEL_H
#define MULTITASKVIEWMODEL_H

#include <QObject>
#include <QIcon>
#include <QList>
#include <QMap>
#include <QRect>
#include <QVariantList>
#include <QString>
#include <kwineffects.h>
#include "multitaskviewmodelhandler.h"

using namespace KWin;
class QGSettings;

class Window
{
    Q_GADGET
    Q_PROPERTY(int id READ id)
    Q_PROPERTY(QString caption READ caption)
    Q_PROPERTY(int x READ x)
    Q_PROPERTY(int y READ y)
    Q_PROPERTY(int width READ width)
    Q_PROPERTY(int height READ height)
    Q_PROPERTY(bool minimized READ minimized)
    Q_PROPERTY(QString uuid READ uuid)

public:
    Window() = default;
    Window(int id, const QString& caption, int x, int y, int width, int height, bool minimized, QString uuid)
        : m_id(id),
          m_caption(caption),
          m_x(x),
          m_y(y),
          m_width(width),
          m_height(height),
          m_isMinimized(minimized),
          m_uuid(uuid) {}
    int id() const { return m_id; }
    QString uuid() const { return m_uuid; }
    const QString& caption() const { return m_caption; }
    int x() const { return m_x; }
    int y() const { return m_y; }
    int width() const { return m_width; }
    int height() const { return m_height; }
    bool minimized() const { return m_isMinimized; }

private:
    int m_id = 0;
    QString m_uuid;
    QString m_caption;
    int m_x = 0;
    int m_y = 0;
    int m_width = 0;
    int m_height = 0;
    bool m_isMinimized = false;
};


class Screen
{
    Q_GADGET
    Q_PROPERTY(int index READ index)
    Q_PROPERTY(QList<QVariant> windowList READ windowList)

public:
    Screen() = default;
    Screen(int index, const QList<QVariant>& windowList)
        : m_index(index),
          m_windowList(windowList) {}

    int index() const { return m_index; }
    const QList<QVariant>& windowList() const { return m_windowList; }

private:
    int m_index = 0;
    QList<QVariant> m_windowList;
};

class Desktop
{
    Q_GADGET
    Q_PROPERTY(int index READ index)
    Q_PROPERTY(QList<QVariant> screenList READ screenList)

public:
    Desktop() = default;
    Desktop(int index, const QList<QVariant>& screenList)
        : m_index(index),
          m_screenList(screenList) {}

    int index() const { return m_index; }
    const QList<QVariant>& screenList() const { return m_screenList; }

private:
    int m_index = 1;
    QList<QVariant> m_screenList;
};

class MultitaskViewModel : public QObject
{
    Q_OBJECT
    Q_PROPERTY(int logicalScreenCount READ logicalScreenCount NOTIFY screenCountChanged)
    Q_PROPERTY(bool isTabletMode READ isTabletMode NOTIFY tabletModeChanged)
    Q_PROPERTY(bool inputMode READ isInputMode NOTIFY inputModeChanged)
public:
    explicit MultitaskViewModel(QObject *parent = nullptr);
    ~MultitaskViewModel();
    Q_INVOKABLE bool currentDesktopHasWindow() const;
    Q_INVOKABLE QList<QVariant> getAppWindowList(int desktopIndex, int screenIndex) const;
    Q_INVOKABLE QList<QVariant> getAllAppWindowList() const;
    Q_INVOKABLE void removeAppWindow(QString windowId);
    Q_INVOKABLE void removeAllAppWindowOnDesktop(int desktopIndex, int screenIndex);
    Q_INVOKABLE void moveAppWindowToDesktop(QString windowId, int srcDesktopInex, int destDesktopInex) const;
    Q_INVOKABLE void activateAppWindow(QString windowId);
    Q_INVOKABLE QIcon getWindowIcon(QString windowId);
    Q_INVOKABLE QList<QVariant> getDesktopList() const;
    Q_INVOKABLE void removeDesktop(int desktopIndex);
    Q_INVOKABLE void appendDesktop();
    Q_INVOKABLE void moveDesktop(int srcDesktopIndex, int destDesktopIndex);
    Q_INVOKABLE int getCurrentDesktop() const;
    Q_INVOKABLE bool isCurrentDesktopShowingDesktop() const;
    Q_INVOKABLE void setCurrentDesktop(int desktopIndex);
    Q_INVOKABLE QRect getFullScreenGeometry(int desktopIndex, int screenIndex) const;
    Q_INVOKABLE bool hasModalChildWindow(QString windowId);
    Q_INVOKABLE bool isTabletMode() const;
    Q_INVOKABLE bool isInputMode() const { return m_inputMode; }
    Q_INVOKABLE QString getCurrentLanguage() const;
    Q_INVOKABLE QString getUKUIStyleName() const;
    Q_INVOKABLE double getUKUIFontSize() const;
    Q_INVOKABLE QString getUKUIFont() const;
    Q_INVOKABLE int getRadius(QString type) const;
    Q_INVOKABLE int getAllWindowCount() const;
    Q_INVOKABLE QString getDesktopUuid(int desktopIndex);
    Q_INVOKABLE QString getOutputUuid(int outputIndex);
    void setIsOpenGLBackend(bool isOpenGLBackend);

signals:
    void appWindowCreated();
    void appWindowRemoved(QString windowId);
    void appWindowActivated();
    void appWindowDesktopChanged(QString windowId, int oldDesktopIndex, int newDesktopIndex);
    void desktopRemoved(int desktopIndex);
    void desktopAppended(int desktopIndex);
    void desktopMoved(int oldDesktopIndex, int newDesktopIndex);
    void currentDesktopChanged(int oldCurrentDesktopIndex, int newCurrentDesktopIndex);
    void screenCountChanged();
    void screenSizeChanged();
    void tabletModeChanged(bool);
    void inputModeChanged(bool);

private slots:
    void onWindowAdded();
    void onWindowDeleted(QString windowId);
    void onWindowActivated();
    void onDesktopPresenceChanged(QString windowId, int oldDesktopIndex, int newDesktopIndex);
    void onCurrentDesktopChanged(int oldCurrentDesktopIndex, int newCurrentDesktopIndex);
    void onNumberScreensChanged();
    void inputModeSlot(bool);

private:
    bool isExtensionMode();
    int logicalScreenCount() const;
    void updateModelData();
    void updateInputMode();
    static bool validateDesktopIndex(int desktopIndex);
    bool validateScreenIndex(int screenIndex) const;
    void connectSignals();
    int getUKUIRadius(QString type) const;

private:
    // 虚拟桌面列表
    QList<QVariant> m_desktopList;
    // 屏幕数量
    int m_screenCount = 0;
    bool m_inputMode;
    // 最大虚拟桌面数量为16
    static constexpr int MAX_DESKTOP_COUNT = 16;
    QGSettings* m_settings = nullptr;
    bool m_isOpenGLBackend = true;
    MultitaskViewModelHandler *model_handler;
};

#endif // MULTITASKVIEWMODEL_H
