/*
 * Copyright 2022 KylinSoft Co., Ltd.
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

import QtQuick 2.12
import QtQuick.Controls 2.5
import QtQuick.Layouts 1.3
import QtQuick.Window 2.12
import QtQuick.Controls.Styles 1.4
import "Controller.js" as Controller

FocusScope {
    id : root
    property var appAreaController : new Controller.AppController(multitaskViewModel, root, multitaskViewManager, ["none"])
    property var currentTabIndex: 0

    ScrollView {
        id : scrollArea
        anchors.fill : parent
        property var flowRows : 0
        property var flowYAnimationNeedStart : false
        property ScrollBar vScrollBar: ScrollBar.vertical
        property ListModel windowPosInfoModel: ListModel{}
        property var rowSpacing : isVerticalLayout ? Math.floor(24 * scaleFactor) : Math.floor(10 * scaleFactor)
        property var columnSpacing : isVerticalLayout ? Math.floor(24 * scaleFactor) : Math.floor(10 * scaleFactor)
        property var tabBorderVisble : false
        contentWidth : parent.width
        contentHeight : getContentHeight()
        ScrollBar.horizontal.policy : ScrollBar.AlwaysOff
        ScrollBar.vertical.policy : isVerticalLayout ? ScrollBar.AlwaysOff : ScrollBar.AsNeeded
        clip : true

        function getContentHeight()
        {
            var currentContentHeight = scrollViewYStartPoint + flowRows * (previewWindowHeight + columnSpacing) - columnSpacing
            if (currentContentHeight > scrollArea.height) {
                return currentContentHeight
            } else {
                return scrollArea.height
            }
        }

        function getFlowY() {
            var flowHeight = scrollArea.flowRows * (previewWindowHeight + scrollArea.columnSpacing) - scrollArea.columnSpacing
            return Math.max((scrollArea.height - flowHeight) / 2, scrollViewYStartPoint)
        }

        onFlowRowsChanged : {
            contentHeight = getContentHeight()
            //当多任务视图打开或者切换工作区时，同样会有flowRowsChanged信号发出
            //这种情况下直接设置坐标即可，不需要播放居中动画
            if (flowYAnimationNeedStart) {
                flowYChangeAnimation.stop()
                flowYChangeAnimation.to = getFlowY()
                flowYChangeAnimation.start()
                flowYAnimationNeedStart = false
            } else {
                flowLayout.y = getFlowY()
            }
        }

        MouseArea{
            anchors.fill : parent
            onClicked : {
                appAreaController.closeMultitaskView()
            }
        }


        NumberAnimation {
            id : scrollMoveAnimation
            target : scrollArea.vScrollBar
            property : "position"
            from : scrollArea.vScrollBar.position
            duration : 300
            easing.type : Easing.InOutQuart
        }

        NumberAnimation {
            id : flowYChangeAnimation
            target : flowLayout
            property : "y"
            from : flowLayout.y
            duration : 300
            easing.type : Easing.InOutQuart
        }

        Flow {
            id : flowLayout
            spacing : scrollArea.rowSpacing
            width : parent.width
            flow : Flow.LeftToRight
            layoutDirection: isLayoutLeftToRight ? Qt.LeftToRight : Qt.RightToLeft
            leftPadding : scrollViewLeftSpace
            rightPadding : scrollViewLeftSpace

            move :Transition { //子项增加时或者减少时的过度，也就是子项的visible变化的过度
                NumberAnimation {
                    properties : "x,y";
                    easing.type : Easing.InOutQuad;
                    duration : isVerticalLayout ? 500 : 300
                }
            }

            Repeater {
                property int loadedCount : 0
                id : previewWindowRepeater
                model : appWindowModel
                activeFocusOnTab: appWindowModel.count !== 0 ? true : false
                focus : true

                Rectangle {
                    id : thumbnailitem
                    height : previewWindowHeight
                    focus: true
                    color : "transparent"
                    property var thumbWidth : getThumbnailSize()[0]
                    property var thumbHeight : getThumbnailSize()[1]
                    property var trueWidth : Math.max(thumbWidth + 2 * appThumbnailMargin, Math.floor(root.width / 10))

                    states: [
                        State {
                            name : "tabActive"
                            when: previewWindowRepeater.activeFocus && index === currentTabIndex && (multitaskViewManager.shouldShowTabBorder() || scrollArea.tabBorderVisble)
                            PropertyChanges {
                                target: appWindow.borderRectangle
                                border.color : tabFocusBorderColor
                            }
                        }
                    ]

                    function getThumbnailSize()
                    {
                        var thumbHeight = thumbRectangleHeight
                        var max_width = root.width / 3
                        var clientWidth = model.width
                        var clientHeight = model.height
                        var thumbWidth = Math.floor(thumbHeight * clientWidth / clientHeight)
                        if (thumbWidth > max_width) {
                            thumbWidth = max_width
                            thumbHeight = Math.floor(thumbWidth * clientHeight / clientWidth)
                        }
                        return [thumbWidth, thumbHeight]
                    }

                    AppPreviewWindow {
                        id : appWindow
                        // Warning: 因为 appWindow 的 parent 会改变，所以不要使用 parent 获取宽高
                        x : isLayoutLeftToRight ? thumbnailitem.width - thumbnailitem.trueWidth : 0
                        width : thumbnailitem.trueWidth
                        height : thumbnailitem.height
                        thumbHeight : thumbnailitem.thumbHeight
                        thumbWidth : thumbnailitem.thumbWidth
                        windowType : "app"
                        x_wid : model.id
                        wid : model.uuid
                        caption : model.caption

                        Drag.onActiveChanged : {
                            if (!Drag.active) {
                                appAreaController.onSourceDropped(appWindow, rootWindow.highlightDesktop, rootWindow.highlightDesktop)
                            } else {
                                //拖拽动画开始前,修改 appWindow 父对象
                                appWindow.parent = mainView
                                zoomOutAnimation()
                            }
                        }
                    }
                }//Rectangle

                onActiveFocusChanged: {
                    if (previewWindowRepeater.activeFocus) {
                        currentTabIndex = 0
                        if (scrollArea.vScrollBar.position > 0) {
                            scrollMoveAnimation.to = 0
                            scrollMoveAnimation.start()
                        }
                        multitaskViewManager.activeAreaChangeDataCollect("Windows Preview Area")
                    }
                }

                /*
                  窗口区域布局居中的问题：
                    组件自身不能实现布局居中，所以使用算法的方式进行计算每一个窗口的大小
                    思路如下：
                        每一行的第一个和最后一个都是一个很长的Rectangle，缩略窗口（trueWidth）只是计算好应该放在哪个位置，
                        其他的窗口都是实际窗口大小（trueWidth），通过spacing就能等间隔的放置在相应的位置.
                        通过计算一行（lineWidth）是否可以在放下下一个窗口，
                        即当前放置的窗口所占的宽度（nowlineWidth）是否大于屏幕的宽度（lineWidth），来确定是否切换行
                */
                function adjustSize()
                {
                    if (loadedCount !== previewWindowCount) {
                        return;
                    }
                    if (previewWindowCount === 0) {
                        //flowRows为0，说明发生了工作区切换或者所有窗口关闭
                        //这两种情况都不需要播放居中动画，所以flowYAnimationNeedStart保持为false
                        scrollArea.flowRows = 0
                        return
                    }

                    var cols = 0;
                    var lineWidth = root.width;
                    var nowlineWidth = 0
                    var nowflowRows = 0
                    var rowWindowNum = 0                  //窗口区域每行有多少窗口
                    var leftX = 0
                    var rightX = 0
                    scrollArea.windowPosInfoModel.clear()

                    // 调整预览窗口外层矩形框的 width，并更新每行第一个预览窗口的 width
                    // 每行第一个预览窗口的 width 和 trueWitdh 不一致，为的是将每一行预览窗口进行居中
                    // 对于所有预览窗口 width >= trueWidth
                    for (var i = 0; i < loadedCount; i++) {
                        var lastline = true
                        // 单独处理每一行的第一个预览窗口，因为第一个预览窗口前面没有 rowSpacing，不应该计算进去
                        nowlineWidth = itemAt(i).trueWidth
                        cols = i
                        for(var j = i + 1; j < loadedCount; j++) {
                            nowlineWidth = nowlineWidth + scrollArea.rowSpacing + itemAt(j).trueWidth
                            if (nowlineWidth < lineWidth - 2 * scrollViewLeftSpace) {
                                itemAt(j).width = itemAt(j).trueWidth
                                cols = j
                            } else {
                                lastline = false
                                nowlineWidth = nowlineWidth - scrollArea.rowSpacing - itemAt(j).trueWidth
                                break;
                            }
                        }
                        itemAt(i).width = itemAt(i).trueWidth + Math.floor((lineWidth - 2 * scrollViewLeftSpace - nowlineWidth) / 2)
                        i = cols
                        nowflowRows++
                    }
                    if (scrollArea.flowRows > nowflowRows) {
                        //行数减少时，重新计算y坐标，播放居中动画
                        scrollArea.flowYAnimationNeedStart = true
                    }
                    scrollArea.flowRows = nowflowRows
                    // 更新缩略图在 Flow 布局中的 X 坐标并存入局部 model
                    var winModel = []
                    for (var m = 0; m < loadedCount; m++) {
                        if (itemAt(m).width !== itemAt(m).trueWidth) {
                            nowlineWidth = itemAt(m).width
                            if (winModel.length > 0) {
                                scrollArea.windowPosInfoModel.append({model : winModel, rowWindowCount : rowWindowNum});
                            }
                            rowWindowNum = 1
                            winModel = []
                        } else {
                            nowlineWidth = nowlineWidth + scrollArea.rowSpacing + itemAt(m).trueWidth
                            rowWindowNum++
                        }
                        leftX = nowlineWidth - itemAt(m).trueWidth
                        rightX = nowlineWidth
                        winModel.push({index: m, leftX: leftX, rightX: rightX, width: itemAt(m).width})
                    }
                    if (winModel.length > 0) {
                        scrollArea.windowPosInfoModel.append({model : winModel, rowWindowCount : rowWindowNum});
                        winModel = []
                    }
                }

                onItemAdded : {
                    loadedCount++;
                    adjustSize();
                }

                onItemRemoved : {
                    loadedCount--;
                    adjustSize();
                }

                Keys.onPressed: {
                    if (event.key === Qt.Key_Left) {
                        scrollArea.tabBorderVisble = true
                        if (currentTabIndex > 0) {
                            currentTabIndex = currentTabIndex - 1
                            var currentNextY = itemAt(currentTabIndex).mapToItem(scrollArea, 0, 0).y
                            if (currentNextY < 0 ) {
                                scrollMoveAnimation.to = scrollArea.vScrollBar.position - (-currentNextY) / scrollArea.contentHeight
                                scrollMoveAnimation.start()
                            }
                        } else if (currentTabIndex === 0) {
                            currentTabIndex = appWindowModel.count - 1
                            var currentNextY = itemAt(currentTabIndex).mapToItem(scrollArea, 0, 0).y
                            if (currentNextY + previewWindowHeight > scrollArea.height) {
                                scrollMoveAnimation.to = scrollArea.vScrollBar.position + (currentNextY + previewWindowHeight - scrollArea.height) / scrollArea.contentHeight
                                scrollMoveAnimation.start()
                            }
                        }
                        event.accepted = true
                    } else if (event.key === Qt.Key_Right ) {
                        scrollArea.tabBorderVisble = true
                        if (currentTabIndex < appWindowModel.count - 1) {
                            currentTabIndex = currentTabIndex + 1
                            var currentNextY = previewWindowRepeater.itemAt(currentTabIndex).mapToItem(scrollArea, 0, 0).y
                            if (currentNextY + previewWindowHeight > scrollArea.height) {
                                scrollMoveAnimation.to = scrollArea.vScrollBar.position + (currentNextY + previewWindowHeight - scrollArea.height) / scrollArea.contentHeight
                                scrollMoveAnimation.start()
                            }
                        } else if (currentTabIndex === appWindowModel.count - 1) {
                            currentTabIndex = 0
                            if (scrollArea.vScrollBar.position > 0) {
                                scrollMoveAnimation.to = 0
                                scrollMoveAnimation.start()
                            }
                        }
                        event.accepted = true
                    } else if (event.key === Qt.Key_Up || event.key === Qt.Key_Down) {
                        scrollArea.tabBorderVisble = true
                        var isKeyUp = event.key === Qt.Key_Up
                        currentTabIndex = appAreaController.keyUpDownSelectWindow(isKeyUp, scrollArea.windowPosInfoModel, currentTabIndex)
                        var currentNextY = itemAt(currentTabIndex).mapToItem(scrollArea, 0, 0).y
                        if (currentNextY < -itemAt(currentTabIndex).height) {
                            scrollMoveAnimation.to = 0
                            scrollMoveAnimation.start()
                        } else if (currentNextY > root.height) {
                            scrollMoveAnimation.to = scrollArea.vScrollBar.position + (currentNextY + previewWindowHeight - scrollArea.height) / scrollArea.contentHeight
                            scrollMoveAnimation.start()
                        }
                    } else if (event.key === Qt.Key_Tab) {
                        if (event.modifiers & Qt.MetaModifier) {
                            appAreaController.closeMultitaskView()
                            return
                        }
                        scrollArea.tabBorderVisble = true
                        currentTabIndex = 0
                        event.accepted = false
                    } else if (event.key === Qt.Key_Return || event.key === Qt.Key_Enter || event.key === Qt.Key_Space) {
                        appAreaController.onWindowClicked(appWindowModel.get(currentTabIndex).uuid);
                        event.accepted = true
                    } else if (event.key === Qt.Key_End) {
                        scrollArea.tabBorderVisble = true
                        currentTabIndex = appWindowModel.count - 1
                        var currentNextY = previewWindowRepeater.itemAt(currentTabIndex).mapToItem(scrollArea,0,0).y
                        if (currentNextY + previewWindowHeight > scrollArea.height) {
                            scrollMoveAnimation.to = scrollArea.vScrollBar.position + (currentNextY + previewWindowHeight - scrollArea.height) / scrollArea.contentHeight
                            scrollMoveAnimation.start()
                        }
                        event.accepted = true
                    } else if (event.key === Qt.Key_Home) {
                        scrollArea.tabBorderVisble = true
                        currentTabIndex = 0
                        if (scrollArea.vScrollBar.position > 0) {
                            scrollMoveAnimation.to = 0
                            scrollMoveAnimation.start()
                        }
                        event.accepted = true
                    } else if (event.key === Qt.Key_PageUp || event.key === Qt.Key_PageDown) {
                        scrollArea.tabBorderVisble = true
                        var isKeyPageUp = event.key === Qt.Key_PageUp
                        currentTabIndex = appAreaController.keyPageUpDownSelectWindow(isKeyPageUp, scrollArea.windowPosInfoModel, currentTabIndex)
                        var currentNextY = previewWindowRepeater.itemAt(currentTabIndex).mapToItem(scrollArea, 0, 0).y
                        if (currentNextY < 0 ) {
                            scrollMoveAnimation.to = scrollArea.vScrollBar.position - (-currentNextY) / scrollArea.contentHeight
                            scrollMoveAnimation.start()
                        } else if (currentNextY + previewWindowHeight > scrollArea.height) {
                            scrollMoveAnimation.to = scrollArea.vScrollBar.position + (currentNextY + previewWindowHeight - scrollArea.height) / scrollArea.contentHeight
                            scrollMoveAnimation.start()
                        }
                        event.accepted = true
                    } else if (event.key === Qt.Key_Escape) {
                        appAreaController.closeMultitaskView()
                    }
                }
            }//Repeater
            Component.onCompleted: {
                previewWindowRepeater.forceActiveFocus()
            }
        }//Flow
    }//ScrollView
}
