/*
 * Copyright 2022 KylinSoft Co., Ltd.
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

/*
    每个预览窗口的UI设计
*/
import QtQuick 2.12
import QtQuick.Controls 2.5
import QtQuick.Layouts 1.3
import QtGraphicalEffects 1.0
import "Controller.js" as Controller
import IconTheme 1.0

Rectangle {
    id : root
    color : "transparent"
    /** windowType is the type of PreviewWindow,
    * the options are : none, app, desktop, creation.
    * Default to none.
    */
    property string windowType : "none"

    /** stateList is the list of state,
    * the options are : default, hover, drag, receive.
    * Default to default.
    */
    property var stateList : ["default", "hover", "drag", "receive"];
    state : "default"
    property string wid : ""
    property int x_wid : -1
    property var thumbHeight
    property var thumbWidth
    property real posX : 0
    property real posY : 0
    property double titleOpacity : 1
    property var controller : new Controller.AppController(multitaskViewModel, root, multitaskViewManager, ["none"]);
    property var captionColor : Qt.rgba(255,255,255,1)
    Drag.active : previewMouseArea.drag.active
    Drag.source : root
    Drag.hotSpot.x : root.width / 2
    Drag.hotSpot.y : root.height / 2
    property alias borderRectangle : previewArea
    property alias caption : caption.text

    // 定义缩放比例系数变量
    property double scaleValue : 0.0
    property double windowDragX : 0.0
    property double windowDragY : 0.0
    property double windowOriginalX : 0.0
    property double windowOriginalY : 0.0


    states : [
        State {
            name : "default"
            PropertyChanges {
                target : previewArea
                border.color : root.parent.state == "tabActive" ? tabFocusBorderColor : "transparent"
            }
            PropertyChanges { target : previewWindowCloseButton; visible : rootWindow.inputMode ? false : true }
        },
        State {
            name : "hover"
            PropertyChanges {
                target : previewArea
                border.color : root.parent.state == "tabActive" ? tabFocusBorderColor : hoverBorderColor
            }
            PropertyChanges { target : previewWindowCloseButton; visible : true }
        }
    ]

    MouseArea {
        id : previewMouseArea
        anchors.fill : parent
        hoverEnabled : true
        preventStealing : true
        drag.axis : Drag.XAndYAxis
        drag.target : root
        onEntered : {
            root.state = "hover"
        }
        onExited : {
            root.state = "default"
        }
        onPressed : {
            windowDragX = mouseX
            windowDragY = mouseY

            //拖拽开始时相对于 mainView 的初始位置
            var originalPos = mapToItem(mainView, 0, 0)
            windowOriginalX = originalPos.x
            windowOriginalY = originalPos.y
        }
        onClicked : {
            controller.onWindowClicked(wid);
        }
        onReleased : {
            root.Drag.drop();
        }

        ColumnLayout {
            spacing : spacingWidth
            anchors.centerIn : parent
            Layout.fillHeight : true
            width : parent.width

            RowLayout {
                id : captionRow
                spacing : Math.floor(8 * scaleFactor)
                Layout.fillHeight : true
                Layout.leftMargin : appBorderWidth
                opacity : titleOpacity
                anchors.leftMargin: appBorderWidth + spacingWidth
                anchors.rightMargin: appBorderWidth + spacingWidth
                layoutDirection: isLayoutLeftToRight ? Qt.LeftToRight : Qt.RightToLeft

                IconTheme {
                    id : iconItem
                    width : iconAndCloseButtonHeight
                    height : width
                    icon : multitaskViewModel.getWindowIcon(wid)
                }

                Label {
                    id : caption
                    verticalAlignment: Qt.AlignVCenter
                    horizontalAlignment: isLayoutLeftToRight ? Qt.AlignLeft : Qt.AlignRight
                    elide : isLayoutLeftToRight ? Text.ElideRight : Text.ElideLeft
                    Layout.fillWidth : true
                    text : ""
                    font.pointSize: multitaskViewModel.getUKUIFontSize()
                    font.family : multitaskViewModel.getUKUIFont()
                    color : captionColor
                }

                CloseButton {
                    id : previewWindowCloseButton
                    width : iconAndCloseButtonHeight
                    height : width
                    radius : width / 2
                    windowType : "app"
                    onExited: {
                        previewWindowCloseButton.state = "default"
                        if (!previewMouseArea.containsMouse) {
                            root.state = "default"
                        }
                    }
                    onClicked: {
                        controller.removePreviewWindow(root.wid)
                    }
                }
            }//RowLayout

            Rectangle {
                id : previewArea
                width : root.width
                height : thumbRectangleHeight + 2 * appThumbnailMargin
                color : "transparent"
                border.color : tabFocusBorderColor
                border.width : appBorderWidth
                radius : appRadius !== 0 ? appRadius + spacingWidth + appBorderWidth : 0

                Rectangle {
                    id : appWindowBack
                    anchors.centerIn: parent
                    height : thumbHeight
                    width : thumbWidth
                    radius : appRadius
                    color : Qt.rgba(1, 1, 1, 0.5)
                    clip : true
                    layer.enabled : appRadius !== 0
                    layer.effect : OpacityMask {
                        maskSource : Rectangle {
                            height : thumbHeight
                            width : thumbWidth
                            radius : appRadius
                        }
                    }
                    Loader {
                        id : appWindowLoader
                        property string thumbnailId : wid
                        height : thumbHeight
                        width : thumbWidth
                        anchors.centerIn: parent
                        source: isWlcom ? "ThumbnailWithWlcom.qml" : "ThumbnailWithKwin.qml"
                    }
                }
            }
        }//culumnLayout
    }

    function returnOriginal(x, y) {
        zoomInAnimation()
    }

    function zoomOutAnimation() {
        transformId.origin.x = windowDragX
        transformId.origin.y = windowDragY
        scaleValue = 0.38
        titleOpacity = 0
        scaleDownAnimation.restart()
    }

    function zoomInAnimation() {
        transformId.origin.x = windowDragX
        transformId.origin.y = windowDragY
        scaleValue = 1
        titleOpacity = 1
        scaleUpAnimation.restart()
    }


    transform : Scale{
        id : transformId
    }

    ParallelAnimation{
        id : scaleDownAnimation
        NumberAnimation {
            target : transformId
            property : "xScale"
            to : scaleValue
            duration : 300
            easing.type : Easing.InOutQuart
        }
        NumberAnimation{
            target : transformId
            property : "yScale"
            to :  scaleValue
            duration : 300
            easing.type : Easing.InOutQuart
        }
        NumberAnimation{
            target : captionRow
            property : "opacity"
            to : titleOpacity
            duration : 300
            easing.type : Easing.InOutQuart
        }
    }

    ParallelAnimation{
        id : scaleUpAnimation
        NumberAnimation {
            target : transformId
            property : "xScale"
            to : scaleValue
            duration : 300
            easing.type : Easing.InOutQuart
        }
        NumberAnimation{
            target : transformId
            property : "yScale"
            to :  scaleValue
            duration : 300
            easing.type : Easing.InOutQuart
        }
        NumberAnimation{
            target : captionRow
            property : "opacity"
            to : titleOpacity
            duration : 300
            easing.type : Easing.InOutQuart
        }

        NumberAnimation {
            target: root
            property: "x"
            duration: 300
            to: windowOriginalX
            easing.type: Easing.InOutQuad
        }
        NumberAnimation {
            target: root
            property: "y"
            duration: 300
            to: windowOriginalY
            easing.type: Easing.InOutQuad
        }

        onFinished: {
            //拖拽还原动画结束后，将 appWindow 的父对象改回 thumbnailitem，并重新设置坐标
            appWindow.parent = thumbnailitem
            appWindow.x = thumbnailitem.width - thumbnailitem.trueWidth
            appWindow.y = 0
        }
    }
}
