/*
 * Copyright 2022 KylinSoft Co., Ltd.
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

/*
    此文件是虚拟桌面的预览窗口的内容设置
*/

import QtQuick 2.12
import QtQuick.Controls 2.5
import QtQuick.Layouts 1.3
import QtQuick.Window 2.12
import org.kylin.background 3.0 as KWin
//import org.kde.kwin 3.0 as KWinComponents
import QtGraphicalEffects 1.0
import "Controller.js" as Controller

Item {
    id : root
    width : isVerticalLayout ? desktopRectWidth : desktopRect.width
    height : isVerticalLayout ? desktopRect.height : desktopRectHeight
    activeFocusOnTab: index === 0 ? true : false
    focus : true
    property string windowType : "desktop"
    property var wid : value
    // Using the Drag attached property, any Item can be made a source of drag and drop events within a scene
    Drag.active : previewMouseArea.drag.active
    Drag.hotSpot.x : width / 2
    Drag.hotSpot.y : height / 2
    property bool windowDragActive : previewMouseArea.drag.active
    property bool nowDragActive: false
    property var controller : new Controller.DesktopController(multitaskViewModel, this, multitaskViewManager, ["app","desktop"])
    property real posX : 0
    property real posY : 0

    states: [
        State {
            name : "tabAndDragActive"
            when: root.activeFocus && root.windowDragActive && (multitaskViewManager.shouldShowTabBorder() ||  desktopRect.desktopTabBorderVisble)
            PropertyChanges { target: previewArea; border.color: tabFocusBorderColor }
            ParentChange { target: root; parent: desktopRect }
        },
        State {
            name : "tabActive"
            when: root.activeFocus && (multitaskViewManager.shouldShowTabBorder() || desktopRect.desktopTabBorderVisble)
            PropertyChanges { target: previewArea; border.color: tabFocusBorderColor }
        },
        State {
            name : "dragActive"
            when : nowDragActive
            ParentChange { target: root; parent: desktopRect }
        }
    ]

    Keys.onPressed: {
        if (event.key === Qt.Key_Return || event.key === Qt.Key_Enter || event.key === Qt.Key_Space) {
            controller.setCurrentDesktop(root.wid)
            controller.onWindowClicked(root.wid)
            event.accepted = true
        } else if (event.key === Qt.Key_Escape) {
            controller.closeMultitaskView()
        } else if (event.modifiers & Qt.MetaModifier) {
            if (event.key === Qt.Key_Tab) {
                controller.closeMultitaskView()
            }
        } else if (event.key === Qt.Key_Tab) {
            if (index === 0 ) {
                desktopListView.currentIndex = desktopWindowModel.count - 1
            }
            console.log("visble",desktopRect.desktopTabBorderVisble)
        }
    }

    onActiveFocusChanged: {
        if (index === 0 && root.activeFocus) {
            desktopListView.currentIndex = 0
            var offsetPosition = isVerticalLayout ? desktopListView.contentX : desktopListView.contentY
            if (offsetPosition > 0) {
                contentMoveAnimation.to = 0
                contentMoveAnimation.start()
            }
            desktopRect.desktopTabBorderVisble = true
            multitaskViewManager.activeAreaChangeDataCollect("Workspace Preview Area")
        }
    }

    // 工作区位置移动触发下方信号
    onWindowDragActiveChanged : {
        // 需要保证让 dragRectangle 先恢复原来的 parent 再移动到合适的位置
        // 因为直接使用 windowDragActive 变量并无法实现上述控制顺序，所以
        // 需要额外的变量 nowDragActive 来实现 ParentChange，然后再移动到
        // 合适的位置
        if (nowDragActive != windowDragActive) {
            nowDragActive = windowDragActive
        }
        if (windowDragActive) {
            opacity = 0.75
        } else {
            // 工作区取消拖动后需要返回原来的位置，如果拖拽产
            // 生了工作区的位置交换也没有关系，也不会影响动效
            opacity = 1
            this.posX = 0
            this.posY = 0
            controller.rejectSource(root)
        }
    }

    Item {
        id : desktopDragRect
        width : desktopRectWidth
        height : mainVerticalLayout.height
        anchors.centerIn : parent
        state : "default"
        property ListModel appmodel : desktopWindowModel.get(index).model

        states : [
            State {
                name : "default"
                PropertyChanges {
                    target : previewArea
                    border.color : desktopDragRect.getWorkspaceDefaultBorderColor()
                }
                PropertyChanges { target : closeButton; visible : rootWindow.inputMode ? false : true }
            },
            State {
                name : "hover"
                PropertyChanges {
                    target : previewArea
                    border.color : desktopDragRect.getWorkspaceHoverBorderColor()
                }
                PropertyChanges { target : closeButton; visible : true }
            }
        ]

        MouseArea {
            id : previewMouseArea
            anchors.fill : parent
            hoverEnabled : true
            preventStealing : true
            drag.axis : isVerticalLayout ? Drag.XAxis : Drag.YAxis
            drag.target : root

            onEntered : {
                controller.setHoverStatus();
            }
            onExited : {
                controller.setDefaultStatus()
            }
            onClicked : {
                controller.onWindowClicked(root.wid)
            }

            ColumnLayout {
                id : mainVerticalLayout
                anchors.centerIn : parent
                spacing : spacingWidth
                Layout.fillHeight : true
                width : parent.width

                RowLayout {
                    id : captionRow
                    spacing : Math.floor(8 * scaleFactor)
                    Layout.fillHeight : true
                    Layout.rightMargin: workspaceBorderWidth
                    layoutDirection: isLayoutLeftToRight ? Qt.LeftToRight : Qt.RightToLeft

                    Label {
                        id : caption
                        verticalAlignment: Qt.AlignVCenter
                        horizontalAlignment: isLayoutLeftToRight ? Qt.AlignLeft : Qt.AlignRight
                        elide : isLayoutLeftToRight ? Text.ElideRight : Text.ElideLeft
                        Layout.fillWidth : true
                        Layout.leftMargin : workspaceBorderWidth
                        height : desktopCaptionHeight
                        text : qsTr("Desktop") + " " + root.wid
                        font.pointSize: multitaskViewModel.getUKUIFontSize()
                        font.family : multitaskViewModel.getUKUIFont()
                        color: rootWindow.desktopCaptionColor
                    }

                    Rectangle {
                        id : iconSpace
                        height : iconAndCloseButtonHeight
                        width : 1
                        color : "transparent"
                    }

                    CloseButton {
                        id : closeButton
                        width : iconAndCloseButtonHeight
                        height : iconAndCloseButtonHeight
                        radius : width / 2
                        windowType : "desktop"
                        onClicked : {
                            controller.removePreviewWindow(root.wid)
                        }
                        onExited: {
                            closeButton.state = "default"
                            if (!previewMouseArea.containsMouse) {
                                root.setDefaultStatus()
                            }
                        }
                    }
                }//RowLayout

                Rectangle {
                    id : previewArea
                    width : desktopRectWidth
                    height : desktopRectHeight - desktopCaptionHeight - spacingWidth
                    color : "transparent"
                    border.color : "transparent"
                    border.width : workspaceBorderWidth
                    radius : workspaceRadius !== 0 ? workspaceRadius + workspaceBorderWidth : 0

                    Rectangle {
                        id: workspace
                        anchors.centerIn: parent
                        width : desktopThumbnailWidth
                        height : desktopThumbnailHeight
                        color : "black"
                        clip : true
                        visible : true
                        layer.enabled : workspaceRadius !== 0
                        layer.effect : OpacityMask {
                            maskSource : Rectangle {
                                width : desktopThumbnailWidth
                                height : desktopThumbnailHeight
                                radius : workspaceRadius
                            }
                        }
                        KWin.DesktopBackground {
                            id : desktopThumbnailBackground
                            anchors.centerIn : parent
                            width : desktopThumbnailWidth
                            height : desktopThumbnailHeight
                            blurAlpha : 0
                        }

                        Loader {
                            id : workSpaceThumbnailLoader
                            width : desktopThumbnailWidth
                            height : desktopThumbnailHeight
                            source: isWlcom ? "WorkSpaceThumbnailWithWlcom.qml" : "WorkSpaceThumbnailWithKwin.qml"
                        }

                        Rectangle {
                            id : desktopMaskColor
                            visible : false
                            anchors.fill: parent
                            color : rootWindow.desktopMaskReceiveColor
                        }
                    }
                }
            }//culumnLayout
        }

        // 应用程序预览窗口拖动到虚拟桌面，虚拟桌面区域收到下方信号
        DropArea {
            id : previewDropArea
            anchors.fill : parent
            visible: !root.windowDragActive
            onEntered : {
                if (drag.source.windowType == "app") {
                    desktopMaskColor.visible = true
                }
                if (drag.source.wid !== root.wid) {
                    controller.onSourceEntered(drag.source, drag.source.wid, root.wid);
                }
            }
            onExited : {
                desktopMaskColor.visible = false
            }
            onDropped : {
                desktopMaskColor.visible = false
                controller.onSourceDropped(drop.source, rootWindow.highlightDesktop, root.wid);
                rootWindow.updateDesktopWindowModel()
            }
        }

        function getWorkspaceDefaultBorderColor()
        {
            if (root.activeFocus && (multitaskViewManager.shouldShowTabBorder() || desktopRect.desktopTabBorderVisble)) {
                return tabFocusBorderColor
            } else if (highlightDesktop == root.wid) {
                return hoverBorderColor
            } else {
                return "transparent"
            }
        }

        function getWorkspaceHoverBorderColor()
        {
            if (root.activeFocus && (multitaskViewManager.shouldShowTabBorder() || desktopRect.desktopTabBorderVisble)) {
                return tabFocusBorderColor
            } else {
                return hoverBorderColor
            }
        }
    }

    NumberAnimation on x{
        id : movePreX
        duration : 0
        running : false
    }

    NumberAnimation on y{
        id : movePreY
        duration : 0
        running : false
    }

    function returnOriginal(x,y)
    {
        movePreX.to = 0
        movePreY.to = 0
        movePreX.from = x
        movePreY.from = y
        movePreX.stop()
        movePreY.stop()
        movePreX.start()
        movePreY.start()
    }

    function setDefaultStatus() {
        desktopDragRect.state = "default"
    }

    function setHoverStatus() {
        desktopDragRect.state = "hover"
    }

    Component.onCompleted: {
        if (root.activeFocusOnTab && rootWindow.appWindowModel.count === 0) {
            root.forceActiveFocus()
            desktopRect.desktopTabBorderVisble = false
        }
    }
}
