/*
 * Copyright 2024 KylinSoft Co., Ltd.
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Library General Public License as
 * published by the Free Software Foundation; either version 2, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

import QtQuick 2.0
import QtQuick.Layouts 1.3
import org.kde.plasma.core 2.0 as PlasmaCore
import org.kde.plasma.components 2.0 as PlasmaComponents
import QtQuick.Controls 2.5
import QtGraphicalEffects 1.0
import org.kde.kquickcontrolsaddons 2.0
import "Controller.js" as Controller
import IconTheme 1.0
import org.kylin.thumbnail 1.0 as WlcomThumbnail
//import org.ukui.kwin 2.0 as KWin

// https://techbase.kde.org/Development/Tutorials/KWin/WindowSwitcher
// https://github.com/KDE/kwin/blob/master/tabbox/switcheritem.h
Rectangle {
    id : tabbox
    property var controller : new Controller.AppController(multitaskViewModel, tabbox, multitaskViewManager, ["none"]);
    property var currentIndex : thumbnailGridView.currentIndex
    property var screenGeometry : multitaskViewModel.getFullScreenGeometry(multitaskViewModel.getCurrentDesktop(), 0)
    property var theme : multitaskViewModel.getUKUIStyleName()
    property ListModel model : ListModel{}
    property var isOpenGL : true
    property int maxHeight : screenGeometry.height * 0.8
    property int maxGridColumnsByWidth: Math.floor(screenGeometry.width / thumbnailGridView.cellWidth)
    property var scaleFactor : Math.min(screenGeometry.width, screenGeometry.height) / 1080
    property int gridColumns : maxGridColumnsByWidth
    property int gridRows : Math.ceil(multitaskViewModel.getAllWindowCount() / gridColumns)
    property int optimalWidth : thumbnailGridView.cellWidth * gridColumns
    property int optimalHeight : thumbnailGridView.cellHeight * gridRows
    property int gridViewMargin : 40 * scaleFactor
    property bool isLayoutLeftToRight: Qt.application.layoutDirection === Qt.LeftToRight

    focus : true
    color : (theme === "ukui-white" || theme === "ukui" || theme === "ukui-light")? Qt.rgba(245/255, 245/255, 245/255, 0.45) : Qt.rgba(35/255, 36/255, 38/255, 0.45)
    width : screenGeometry.width
    height : Math.min(optimalHeight + 2 * gridViewMargin, maxHeight)
    clip : true
    Keys.enabled : true
    onActiveFocusChanged : {
        if(!tabbox.activeFocus) {
            multitaskViewManager.close()
        } else {
             /**
             1. wlcom环境特定情况（快速操作按下alt+tab并松开）下，会出现本界面不隐藏问题，原因如下：
             由于wlcom合成器与窗口选择器是两个独立的进程，使得窗口选择器被快捷键拉起时，向合成器发出的GrabKeyboard请求未被及时响应
             导致在特定情况下窗口选择器无法收到Key_Alt的Released信号，进而无法通过响应alt released事件来关闭此界面
             2. 当前实现，此界面显示时可以认为Key_Alt必定处于按下状态，因此针对特定情况处理如下：
             在此界面显示并获取键盘焦点后，立即查询Key_Alt是否处于按下状态
             如果按未处于按下状态，则激活选定窗口并退出；否则，不做特殊处理
             **/
            // fix https://gitee.com/openkylin/ukui-window-switch/issues/IAID8L?from=project-issue
            if (!multitaskViewManager.isAltKeyPressed()) {
                activateSelectedWindow();
            }
        }
    }

    GridView {
        id : thumbnailGridView
        anchors.fill : parent
        anchors.bottomMargin : tabbox.gridViewMargin
        anchors.topMargin : tabbox.gridViewMargin
        anchors.leftMargin : count > tabbox.maxGridColumnsByWidth ? (tabbox.width-tabbox.maxGridColumnsByWidth * cellWidth) / 2 :
                                                                   (tabbox.width-count * cellWidth) / 2
        anchors.rightMargin : count > tabbox.maxGridColumnsByWidth ? (tabbox.width-tabbox.maxGridColumnsByWidth * cellWidth) / 2 :
                                                                    (tabbox.width-count * cellWidth) / 2
        LayoutMirroring.enabled: !isLayoutLeftToRight
        model : tabbox.model
        property int iconSize : 24 * tabbox.scaleFactor
        property int captionRowHeight : 33 * tabbox.scaleFactor * units.devicePixelRatio
        property int thumbnailWidth : (tabbox.screenGeometry.width - 2 * tabbox.gridViewMargin) / 4
        property int thumbnailHeight : tabbox.screenGeometry.height / 22 * 5
        property int spaceWidth : 2 * tabbox.scaleFactor
        property int outerBorderWidth : 4 * tabbox.scaleFactor
        property int innerBorderWidth : 2 * tabbox.scaleFactor
        property int appRadius : 8 * tabbox.scaleFactor
        property int thumbnailOuterRadius : tabbox.isOpenGL ? appRadius + outerBorderWidth + innerBorderWidth + 2 * spaceWidth : 0
        property int thumbnailInterRadius : tabbox.isOpenGL ? appRadius + innerBorderWidth + spaceWidth : 0
        cellWidth : 2 * (innerBorderWidth + outerBorderWidth) + 4 * spaceWidth + thumbnailWidth
        cellHeight : 2 * (innerBorderWidth + outerBorderWidth) + 5 * spaceWidth + captionRowHeight + thumbnailHeight
        keyNavigationWraps : true
        focus : true

        delegate : Rectangle {
            id : thumbnailGridItem
            width : thumbnailGridView.cellWidth
            height : thumbnailGridView.cellHeight
            color : "transparent"
            border.color : (tabbox.theme === "ukui-white" || tabbox.theme === "ukui" || tabbox.theme === "ukui-light") ? "#262626" : "#FFFFFF"
            border.width : index == thumbnailGridView.currentIndex ? thumbnailGridView.outerBorderWidth : 0
            // highlight在快速绘制的时候会出现绘制的坐标偏移的问题，故不使用
            radius : thumbnailGridView.thumbnailOuterRadius

            MouseArea {
                anchors.fill : parent
                onClicked : {
                    thumbnailGridItem.select();
                }
                hoverEnabled : true

                onEntered : {
                    thumbnailBorderItem.border.width = thumbnailGridView.innerBorderWidth;
                }

                onExited : {
                    thumbnailBorderItem.border.width = 0;
                }
            }
            function select() {
                thumbnailGridView.currentIndex = index;
                controller.onWindowClicked(model.uuid);
            }

            function getThumbnailSize()
            {
                var thumbHeight = thumbnailGridView.thumbnailHeight
                var max_width = thumbnailGridView.thumbnailWidth
                var clientWidth = model.width
                var clientHeight = model.height
                var thumbWidth = Math.floor(thumbHeight * clientWidth / clientHeight)
                if (thumbWidth > max_width) {
                    thumbWidth = max_width
                    thumbHeight = Math.floor(thumbWidth * clientHeight / clientWidth)
                }
                return [thumbWidth, thumbHeight]
            }

            ColumnLayout {
                spacing : 0
                anchors.centerIn : parent
                height : parent.height - 2 * (thumbnailGridView.outerBorderWidth + thumbnailGridView.spaceWidth)
                width : parent.width - 2 * (thumbnailGridView.outerBorderWidth + thumbnailGridView.spaceWidth)
                x : thumbnailGridView.outerBorderWidth + thumbnailGridView.spaceWidth
                y : thumbnailGridView.outerBorderWidth + thumbnailGridView.spaceWidth
                RowLayout {
                    id : captionRow
                    spacing : 2 * thumbnailGridView.spaceWidth
                    Layout.fillHeight : true
                    layoutDirection: isLayoutLeftToRight ? Qt.LeftToRight : Qt.RightToLeft

                    IconTheme {
                        id : iconItem
                        width : thumbnailGridView.iconSize
                        height : thumbnailGridView.iconSize
                        icon : multitaskViewModel.getWindowIcon(model.uuid)
                    }

                    Label {
                        id : label
                        verticalAlignment : Qt.AlignVCenter
                        Layout.fillWidth : true
                        color : (tabbox.theme === "ukui-white" || tabbox.theme === "ukui" || tabbox.theme === "ukui-light") ? "#000000" : "#FFFFFF"
                        text : model.caption
                        horizontalAlignment: isLayoutLeftToRight ? Qt.AlignLeft : Qt.AlignRight
                        elide : isLayoutLeftToRight ? Text.ElideRight : Text.ElideLeft
                    }

                    //label高度受字号控制，此Rectangle起占位作用，防止字号改变时缩略图位置改变
                    Rectangle {
                        width : 1
                        height : thumbnailGridView.captionRowHeight
                        color : "transparent"
                    }
                }

                // KWin.ThumbnailItem needs a container
                // otherwise it will be drawn the same size as the parent ColumnLayout
                Rectangle {
                    color : "transparent"
                    Layout.fillHeight : true
                    Layout.fillWidth : true
                    Rectangle {
                        id : thumbnailBorderItem
                        x : (parent.width - thumbnailBorderItem.width) / 2
                        y : thumbnailGridView.spaceWidth
                        width : getThumbnailSize()[0] + 2 * (thumbnailGridView.innerBorderWidth + thumbnailGridView.spaceWidth)
                        height : thumbnailGridView.thumbnailHeight + 2 * (thumbnailGridView.innerBorderWidth + thumbnailGridView.spaceWidth)
                        color : "transparent"
                        border.color : (tabbox.theme === "ukui-white" || tabbox.theme === "ukui" || tabbox.theme === "ukui-light") ? "#262626" : "#FFFFFF"
                        border.width : 0
                        radius : thumbnailGridView.thumbnailInterRadius
                        Rectangle {
                            id : thumbnailBackItem
                            width : getThumbnailSize()[0]
                            height : getThumbnailSize()[1]
                            anchors.centerIn : parent
                            color : "transparent"
                            clip : true
                            layer.enabled : tabbox.isOpenGL ? true : false
                            layer.effect : OpacityMask {
                                maskSource : Rectangle {
                                    width : thumbnailBackItem.width
                                    height : thumbnailBackItem.height
                                    radius : Math.min(thumbnailGridView.appRadius, Math.floor(Math.min(width, height) / 2))
                                }
                            }


                            WlcomThumbnail.ThumbnailItem {
                                anchors.fill: parent
                                windowId: model.uuid
                                visible: thumbnailGridView.visible
                            }
                        }
                    }
                }
            }
        } // GridView.delegate

        Connections {
            target : tabbox
            function onCurrentIndexChanged() {
                thumbnailGridView.currentIndex = tabbox.currentIndex;
            }
        }
    } // GridView

    Keys.onDeletePressed : {
        multitaskViewManager.close()
    }

    Keys.onEscapePressed : {
        multitaskViewManager.close()
    }

    Keys.onBacktabPressed : {
        if (isLayoutLeftToRight) {
            thumbnailGridView.moveCurrentIndexLeft();
        } else {
            thumbnailGridView.moveCurrentIndexRight();
        }
    }

    Keys.onTabPressed : {
        if (isLayoutLeftToRight) {
            thumbnailGridView.moveCurrentIndexRight();
        } else {
            thumbnailGridView.moveCurrentIndexLeft();
        }
    }

    Keys.onPressed : {
        if (event.key === Qt.Key_Left) {
            thumbnailGridView.moveCurrentIndexLeft();
        } else if (event.key === Qt.Key_Right) {
            thumbnailGridView.moveCurrentIndexRight();
        } else if (event.key === Qt.Key_Up) {
            thumbnailGridView.moveCurrentIndexUp();
        } else if (event.key === Qt.Key_Down) {
            thumbnailGridView.moveCurrentIndexDown();
        } else {
            return;
        }
    }

    Keys.onReleased : {
        if (event.key === Qt.Key_Alt) {
            activateSelectedWindow();
        }
    }

    function activateSelectedWindow() {
        controller.onWindowClicked(model.get(thumbnailGridView.currentIndex).uuid);
        thumbnailGridView.visible = false;
    }

    function resetAppWindowModel(deskindex) {
        var windowInfo = multitaskViewModel.getAllAppWindowList()
        model.clear();
        for (var i in windowInfo) {
            var info = windowInfo[i]
            model.append({"x" : info.x, "y" : info.y,
                             "width" : info.width, "height" : info.height,
                             "id" : info.id,"minimized" : info.minimized, "caption" : info.caption,
                             "uuid" : String(info.uuid)});
        }
    }

    Component.onCompleted : {
        resetAppWindowModel(multitaskViewModel.getCurrentDesktop())
        tabbox.currentIndex = model.count > 1 ? 1 : 0
    }
}
